@@ -1,4 +1,4 @@
-// Copyright (c) 2014, Riverbank Computing Limited
+// Copyright (c) 2017, Riverbank Computing Limited
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -36,24 +36,42 @@
 
 #include "frozen_bootstrap.h"
 
+#if PY_VERSION_HEX >= 0x03050000
+#include "frozen_bootstrap_external.h"
+#endif
+
 #if defined(PYQTDEPLOY_FROZEN_MAIN)
 #include "frozen_main.h"
 #endif
 
 
 #if PY_MAJOR_VERSION >= 3
-#define BOOTSTRAP_MODULE    "_frozen_importlib"
-#define PDYTOOLS_INIT       PyInit_pdytools
-#define CONST_CAST(s)       s
+
+#define BOOTSTRAP_MODULE            "_frozen_importlib"
+#define BOOTSTRAP_EXTERNAL_MODULE   "_frozen_importlib_external"
+#define PDYTOOLS_INIT               PyInit_pdytools
+#define CONST_CAST(s)               s
 extern "C" PyObject *PyInit_pdytools(void);
+
+#if defined(Q_OS_WIN)
+#define WIDE_ARGV
+#endif
+
 #else
+
 #define BOOTSTRAP_MODULE    "__bootstrap__"
 #define PDYTOOLS_INIT       initpdytools
 #define CONST_CAST(s)       const_cast<char *>(s)
 extern "C" void initpdytools(void);
+
 #endif
 
 
+// The internal API.
+void pdytools_init_executable_dir(const QString &argv0);
+const QDir &pdytools_get_executable_dir();
+
+
 // We use Qt as the source of the locale information, partly because it
 // officially supports Android.
 static QTextCodec *locale_codec;
@@ -61,22 +79,42 @@ static QTextCodec *locale_codec;
 
 // Foward declarations.
 static int handle_exception();
-static int append_path_dirs(PyObject *list, const char **path_dirs,
-        const QDir &exec_dir);
+static int append_path_dirs(PyObject *list, const char **path_dirs);
 #if PY_MAJOR_VERSION < 3
 static PyObject *string_from_qstring(const QString &qs);
 #endif
 
 
-int pyqtdeploy_start(int argc, char **argv, struct _inittab *extension_modules,
-        const char *main_module, const char *entry_point,
-        const char **path_dirs)
+#if defined(WIDE_ARGV)
+int pyqtdeploy_start(int argc, wchar_t **w_argv,
+        struct _inittab *extension_modules, const char *main_module,
+        const char *entry_point, const char **path_dirs)
+#else
+int pyqtdeploy_start(int argc, char **argv,
+        struct _inittab *extension_modules, const char *main_module,
+        const char *entry_point, const char **path_dirs)
+#endif
 {
     // The replacement table of frozen modules.
     static struct _frozen modules[] = {
-        {CONST_CAST(BOOTSTRAP_MODULE), frozen_pyqtdeploy_bootstrap, sizeof (frozen_pyqtdeploy_bootstrap)},
+        {
+            CONST_CAST(BOOTSTRAP_MODULE),
+            frozen_pyqtdeploy_bootstrap,
+            sizeof (frozen_pyqtdeploy_bootstrap)
+        },
+#if PY_VERSION_HEX >= 0x03050000
+        {
+            CONST_CAST(BOOTSTRAP_EXTERNAL_MODULE),
+            frozen_pyqtdeploy_bootstrap_external,
+            sizeof (frozen_pyqtdeploy_bootstrap_external)
+        },
+#endif
 #if defined(PYQTDEPLOY_FROZEN_MAIN)
-        {CONST_CAST("__main__"), frozen_pyqtdeploy_main, sizeof (frozen_pyqtdeploy_main)},
+        {
+            CONST_CAST("__main__"),
+            frozen_pyqtdeploy_main,
+            sizeof (frozen_pyqtdeploy_main)
+        },
 #endif
         {NULL, NULL, 0}
     };
@@ -86,7 +124,11 @@ int pyqtdeploy_start(int argc, char **argv, struct _inittab *extension_modules,
 
     if (!locale_codec)
     {
+#if defined(WIDE_ARGV)
+        fwprintf(stderr, L"%s: no locale codec found\n", w_argv[0]);
+#else
         fprintf(stderr, "%s: no locale codec found\n", argv[0]);
+#endif
         return 1;
     }
 
@@ -94,6 +136,9 @@ int pyqtdeploy_start(int argc, char **argv, struct _inittab *extension_modules,
     Py_FrozenFlag = 1;
     Py_NoSiteFlag = 1;
     Py_IgnoreEnvironmentFlag = 1;
+#if defined(PYQTDEPLOY_OPTIMIZED)
+    Py_OptimizeFlag = 1;
+#endif
 
 #if PY_MAJOR_VERSION >= 3
     if (!Py_FileSystemDefaultEncoding)
@@ -117,7 +162,11 @@ int pyqtdeploy_start(int argc, char **argv, struct _inittab *extension_modules,
     // Add the importer to the table of builtins.
     if (PyImport_AppendInittab("pdytools", PDYTOOLS_INIT) < 0)
     {
+#if defined(WIDE_ARGV)
+        fwprintf(stderr, L"%s: PyImport_AppendInittab() failed\n", w_argv[0]);
+#else
         fprintf(stderr, "%s: PyImport_AppendInittab() failed\n", argv[0]);
+#endif
         return 1;
     }
 
@@ -125,11 +174,16 @@ int pyqtdeploy_start(int argc, char **argv, struct _inittab *extension_modules,
     if (extension_modules != NULL)
         if (PyImport_ExtendInittab(extension_modules) < 0)
         {
+#if defined(WIDE_ARGV)
+            fwprintf(stderr, L"%s: PyImport_ExtendInittab() failed\n", w_argv[0]);
+#else
             fprintf(stderr, "%s: PyImport_ExtendInittab() failed\n", argv[0]);
+#endif
             return 1;
         }
 
 #if PY_MAJOR_VERSION >= 3
+#if !defined(WIDE_ARGV)
     // Convert the argument list to wide characters using the locale codec.
     wchar_t **w_argv = new wchar_t *[argc + 1];
 
@@ -145,6 +199,7 @@ int pyqtdeploy_start(int argc, char **argv, struct _inittab *extension_modules,
     }
 
     w_argv[argc] = NULL;
+#endif
 
     // Initialise the Python v3 interpreter.
     Py_SetProgramName(w_argv[0]);
@@ -162,9 +217,16 @@ int pyqtdeploy_start(int argc, char **argv, struct _inittab *extension_modules,
 #endif
 
     // Set sys.frozen.
-    if (PySys_SetObject("frozen", Py_True) < 0)
+    if (PySys_SetObject(CONST_CAST("frozen"), Py_True) < 0)
         return handle_exception();
 
+    // Initialise the directory containing the executable.
+#if PY_MAJOR_VERSION >= 3
+    pdytools_init_executable_dir(QString::fromWCharArray(w_argv[0]));
+#else
+    pdytools_init_executable_dir(locale_codec->toUnicode(argv[0]));
+#endif
+
     // Configure sys.path.
     if (path_dirs != NULL)
     {
@@ -172,12 +234,7 @@ int pyqtdeploy_start(int argc, char **argv, struct _inittab *extension_modules,
 
         if (py_path)
         {
-            // Get the directory containing the executable.
-            QDir exec_dir(locale_codec->toUnicode(argv[0]));
-            exec_dir.makeAbsolute();
-            exec_dir.cdUp();
-
-            if (append_path_dirs(py_path, path_dirs, exec_dir) < 0)
+            if (append_path_dirs(py_path, path_dirs) < 0)
                 return handle_exception();
         }
     }
@@ -228,7 +285,7 @@ int pyqtdeploy_start(int argc, char **argv, struct _inittab *extension_modules,
         return handle_exception();
 
     // Call the entry point.
-    if (!PyObject_CallMethod(mod, entry_point, NULL))
+    if (!PyObject_CallMethod(mod, CONST_CAST(entry_point), NULL))
         return handle_exception();
 #endif
 
@@ -250,6 +307,7 @@ static int handle_exception()
         PyObject *exc, *value, *tb;
 
         PyErr_Fetch(&exc, &value, &tb);
+        PyErr_NormalizeException(&exc, &value, &tb);
 
         if (!value || value == Py_None)
         {
@@ -261,14 +319,19 @@ static int handle_exception()
 
             if (code)
             {
+                if (code == Py_None)
+                {
+                    exit_code = 0;
+                    Py_DECREF(code);
+                }
 #if PY_MAJOR_VERSION >= 3
-                if (PyLong_Check(code))
+                else if (PyLong_Check(code))
                 {
                     exit_code = (int)PyLong_AsLong(code);
                     Py_DECREF(code);
                 }
 #else
-                if (PyInt_Check(code))
+                else if (PyInt_Check(code))
                 {
                     exit_code = (int)PyInt_AsLong(code);
                     Py_DECREF(code);
@@ -302,8 +365,7 @@ static int handle_exception()
 
 // Extend a list with an array of UTF-8 encoded path directory names.  Return
 // -1 if there was an error.
-static int append_path_dirs(PyObject *list, const char **path_dirs,
-        const QDir &exec_dir)
+static int append_path_dirs(PyObject *list, const char **path_dirs)
 {
     const char *path_dir_utf8;
 
@@ -328,11 +390,16 @@ static int append_path_dirs(PyObject *list, const char **path_dirs,
 
         // Make sure the path is absolute.
         if (QDir::isRelativePath(path_dir))
+        {
+            const QDir &exec_dir = pdytools_get_executable_dir();
             path_dir = exec_dir.filePath(path_dir);
+        }
 
-        // Convert to the native format.  (Note that we don't resolve symbolic
-        // links.
-        path_dir = QDir::toNativeSeparators(QDir::cleanPath(path_dir));
+        // Convert to the native format unless it is a resource path.  (Note
+        // that we don't resolve symbolic links.)
+        path_dir = QDir::cleanPath(path_dir);
+        if (!path_dir.startsWith(QChar(':')))
+            path_dir = QDir::toNativeSeparators(path_dir);
 
         // Convert to a Python string.
         PyObject *py_path_dir;
