@@ -1,4 +1,4 @@
-# Copyright (c) 2014, Riverbank Computing Limited
+# Copyright (c) 2017, Riverbank Computing Limited
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -25,12 +25,13 @@
 
 
 from PyQt5.QtCore import pyqtSignal, Qt
-from PyQt5.QtWidgets import (QButtonGroup, QCheckBox, QFileDialog, QGridLayout,
-        QGroupBox, QLineEdit, QRadioButton, QVBoxLayout, QWidget)
+from PyQt5.QtWidgets import (QButtonGroup, QCheckBox, QComboBox, QFileDialog,
+        QGridLayout, QGroupBox, QHBoxLayout, QLineEdit, QRadioButton, QWidget)
 
+from ..metadata import supported_python_versions
 from .better_form import BetterForm
 from .filename_editor import FilenameEditor
-from .qrc_package_editor import QrcPackageEditor
+from .package_editor import PackageEditor
 
 
 class ApplicationPage(QWidget):
@@ -42,6 +43,9 @@ class ApplicationPage(QWidget):
     # Emitted when the user changes the PyQt version.
     pyqt_version_changed = pyqtSignal(bool)
 
+    # Emitted when the user changes the Python target version.
+    python_target_version_changed = pyqtSignal()
+
     @property
     def project(self):
         """ The project property getter. """
@@ -94,34 +98,46 @@ class ApplicationPage(QWidget):
 
         self._sys_path_edit = QLineEdit(
                 placeholderText="Additional sys.path directories",
-                whatsThis="A space separated list of additional directories "
-                        "to add to <tt>sys.path</tt>. Only set this if you "
-                        "want to allow external packages to be imported.",
+                whatsThis="A space separated list of additional directories, "
+                        "ZIP files and eggs to add to <tt>sys.path</tt>. Only "
+                        "set this if you want to allow external packages to "
+                        "be imported.",
                 textEdited=self._sys_path_changed)
         form.addRow("sys.path", self._sys_path_edit)
 
         layout.addLayout(form, 0, 0)
 
-        options_layout = QVBoxLayout()
-        self._pyqt_versions_bg = QButtonGroup()
+        options_layout = BetterForm()
 
-        for version in ('PyQt5', 'PyQt4'):
-            rb = QRadioButton(version)
-            options_layout.addWidget(rb)
-            self._pyqt_versions_bg.addButton(rb)
+        self._py_version_edit = QComboBox(
+                whatsThis="Select the target Python version.")
+        for major, minor, patch in supported_python_versions:
+            self._py_version_edit.addItem(
+                    "v{0}.{1}.{2}".format(major, minor, patch))
+        self._py_version_edit.currentIndexChanged.connect(
+                self._py_version_changed)
+        options_layout.addRow("Target Python version", self._py_version_edit)
 
-        self._pyqt_versions_bg.buttonToggled.connect(
+        self._pyqt_version_edit = QComboBox(
+                whatsThis="Select the PyQt version.")
+        self._pyqt_version_edit.addItems(["PyQt4", "PyQt5"])
+        self._pyqt_version_edit.currentIndexChanged.connect(
                 self._pyqt_version_changed)
+        options_layout.addRow("Target PyQt version", self._pyqt_version_edit)
 
         self._console_edit = QCheckBox("Use console (Windows)",
+                whatsThis="Enable console output for Windows applications. "
+                        "Console output will be enabled automatically if no "
+                        "graphical PyQt modules are used.",
                 stateChanged=self._console_changed)
-        options_layout.addWidget(self._console_edit)
+        options_layout.addRow(self._console_edit)
 
-        self._bundle_edit = QCheckBox("Application bundle (OS/X)",
+        self._bundle_edit = QCheckBox("Application bundle (macOS)",
+                whatsThis="Build an application bundle on macOS. If it is not "
+                        "checked then the application will be built as a "
+                        "simple executable.",
                 stateChanged=self._bundle_changed)
-        options_layout.addWidget(self._bundle_edit)
-
-        options_layout.addStretch()
+        options_layout.addRow(self._bundle_edit)
 
         layout.addLayout(options_layout, 0, 1)
 
@@ -145,15 +161,15 @@ class ApplicationPage(QWidget):
         self._sys_path_edit.setText(project.sys_path)
         self._package_edit.configure(project.application_package, project)
 
-        blocked = self._pyqt_versions_bg.blockSignals(True)
+        blocked = self._py_version_edit.blockSignals(True)
+        self._py_version_edit.setCurrentIndex(
+                supported_python_versions.index(project.python_target_version))
+        self._py_version_edit.blockSignals(blocked)
 
-        for rb in self._pyqt_versions_bg.buttons():
-            if rb.text() == 'PyQt5':
-                rb.setChecked(project.application_is_pyqt5)
-            else:
-                rb.setChecked(not project.application_is_pyqt5)
-
-        self._pyqt_versions_bg.blockSignals(blocked)
+        blocked = self._pyqt_version_edit.blockSignals(True)
+        self._pyqt_version_edit.setCurrentIndex(
+                1 if project.application_is_pyqt5 else 0)
+        self._pyqt_version_edit.blockSignals(blocked)
 
         blocked = self._console_edit.blockSignals(True)
         self._console_edit.setCheckState(
@@ -165,14 +181,22 @@ class ApplicationPage(QWidget):
                 Qt.Checked if project.application_is_bundle else Qt.Unchecked)
         self._bundle_edit.blockSignals(blocked)
 
-    def _pyqt_version_changed(self, button, checked):
+    def _py_version_changed(self, idx):
+        """ Invoked when the user changes the Python version number. """
+
+        self.project.python_target_version = supported_python_versions[idx]
+        self.project.modified = True
+
+        self.python_target_version_changed.emit()
+
+    def _pyqt_version_changed(self, idx):
         """ Invoked when the user changes the PyQt version number. """
 
-        if button.text() == 'PyQt5':
-            self.project.application_is_pyqt5 = checked
-            self.project.modified = True
+        pyqt5 = (idx == 1)
+        self.project.application_is_pyqt5 = pyqt5
+        self.project.modified = True
 
-            self.pyqt_version_changed.emit(checked)
+        self.pyqt_version_changed.emit(pyqt5)
 
     def _console_changed(self, state):
         """ Invoked when the user changes the console state. """
@@ -216,7 +240,7 @@ class ApplicationPage(QWidget):
         self.project.modified = True
 
 
-class _ApplicationPackageEditor(QrcPackageEditor):
+class _ApplicationPackageEditor(PackageEditor):
     """ A memory filesystem package editor for the application package. """
 
     # The editor title.
@@ -225,7 +249,13 @@ class _ApplicationPackageEditor(QrcPackageEditor):
     def __init__(self):
         """ Initialise the editor. """
 
-        super().__init__(show_root=True, scan="Scan...")
+        super().__init__(show_root=True, scan="Scan...",
+                scan_whats_this="Display a dialog from which you can select "
+                        "the application package directory.",
+                whats_this="This shows the scanned application package "
+                        "directory and its contents. Check those directories "
+                        "and files that should be included in the "
+                        "application.")
 
         self._project = None
 
@@ -235,8 +265,8 @@ class _ApplicationPackageEditor(QrcPackageEditor):
         project = self._project
         application_package = project.application_package
 
-        orig = default = application_package.name
-        if default != '':
+        default = application_package.name
+        if default is not None:
             default = project.path_from_user(default)
 
         root = QFileDialog.getExistingDirectory(self.parentWidget(),
