@@ -1,4 +1,4 @@
-# Copyright (c) 2014, Riverbank Computing Limited
+# Copyright (c) 2017, Riverbank Computing Limited
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -29,21 +29,20 @@ from xml.etree.ElementTree import Element, ElementTree, SubElement
 
 from PyQt5.QtCore import QDir, QFileInfo, QObject, pyqtSignal
 
-from ..metadata import get_latest_supported_python_version, get_python_metadata
+from ..metadata import get_python_metadata, supported_python_versions
+from ..platforms import Platform
 from ..user_exception import UserException
 
 
 class Project(QObject):
     """ The encapsulation of a project. """
 
-    # The current project version.  Versions 0 to 3 are no longer supported.
-    # Note that we added the 'patch' attribute to the 'Project/Python' element
-    # to Version 4 without bumping the version number to 5.  This is because we
-    # haven't got round to implementing a proper mechanism to update the
-    # format.  It will only be a problem if a project targets Python v3.4.2 or
-    # later (using pyqtdeploy v0.7 or later) and is then edited with pyqtdeploy
-    # v0.6.
-    version = 4
+    # The minimum supported project version.  At the moment a project will be
+    # automatically updated to the current version when saved.
+    min_version = 4
+
+    # The current project version.
+    version = 7
 
     # Emitted when the modification state of the project changes.
     modified_changed = pyqtSignal(bool)
@@ -97,22 +96,30 @@ class Project(QObject):
         self.application_package = QrcPackage()
         self.application_script = ''
         self.application_entry_point = ''
-        self.build_dir = 'build'
-        self.external_libraries = []
+        self.external_libraries = {}
         self.other_extension_modules = []
         self.other_packages = []
         self.pyqt_modules = []
-        self.python_host_interpreter = ''
-        self.python_source_dir = ''
-        self.python_ssl = False
-        self.python_target_include_dir = ''
-        self.python_target_library = ''
-        self.python_target_stdlib_dir = ''
-        self.python_target_version = get_latest_supported_python_version()
-        self.qmake = ''
+        self.python_use_platform = ['win']
+        self.python_target_version = supported_python_versions[0]
+        self.qmake_configuration = ''
         self.standard_library = []
         self.sys_path = ''
 
+        self.set_default_locations()
+
+    def set_default_locations(self):
+        """ Set the various locations to their default values. """
+
+        self.python_host_interpreter = '$SYSROOT/host/bin/python'
+
+        self.python_source_dir = '$SYSROOT/src/Python-$PDY_PY_MAJOR.$PDY_PY_MINOR.$PDY_PY_MICRO'
+        self.python_target_include_dir = '$SYSROOT/include/python$PDY_PY_MAJOR.$PDY_PY_MINOR'
+        self.python_target_library = '$SYSROOT/lib/libpython$PDY_PY_MAJOR.$PDY_PY_MINOR.a'
+        self.python_target_stdlib_dir = '$SYSROOT/lib/python$PDY_PY_MAJOR.$PDY_PY_MINOR'
+
+        self.using_default_locations = True
+
     def path_to_user(self, path):
         """ Convert a file name to one that is relative to the project name if
         possible and uses native separators.
@@ -158,15 +165,34 @@ class Project(QObject):
 
         return self._fileinfo_from_user(name).completeBaseName()
 
+    def expandvars(self, path):
+        """ Call os.path.expandvars() after expanding some internal values. """
+
+        major, minor, micro = self.python_target_version
+        major = str(major)
+        minor = str(minor)
+        micro = str(micro)
+
+        path = path.replace('$PDY_PY_MAJOR', major)
+        path = path.replace('${PDY_PY_MAJOR}', major)
+
+        path = path.replace('$PDY_PY_MINOR', minor)
+        path = path.replace('${PDY_PY_MINOR}', minor)
+
+        path = path.replace('$PDY_PY_MICRO', micro)
+        path = path.replace('${PDY_PY_MICRO}', micro)
+
+        return os.path.expandvars(path)
+
     def _fileinfo_from_user(self, user_path):
         """ Convert the name of a file or directory specified by the user to a
         QFileInfo instance.  A user path may be relative to the name of the
         project and may contain environment variables.
         """
 
-        fi = QFileInfo(os.path.expandvars(user_path.strip()))
+        fi = QFileInfo(self.expandvars(user_path.strip()))
 
-        if fi.isRelative():
+        if fi.isRelative() and self._name is not None:
             fi = QFileInfo(self._name.canonicalPath() + '/' + fi.filePath())
 
         return fi
@@ -224,9 +250,10 @@ class Project(QObject):
 
         dep_state.visit = visit
 
-        if dep_state.module.ssl is not None:
-            if dep_state.module.ssl != self.python_ssl:
-                return
+        if dep_state.module.builtin:
+            # This will mean that the explicit and implicit states will remain
+            # None and so the module will be omitted from the list.
+            return
 
         dep_state.explicit = (name in self.standard_library)
 
@@ -234,6 +261,20 @@ class Project(QObject):
             dep_state.implicit = True
 
         for dep in dep_state.module.deps:
+            # If the first character of the module is '?' then it should be
+            # excluded if SSL support is disabled.  If the first character is
+            # '!' then it should be excluded if SSL support is enabled.
+            if dep[0] == '?':
+                if 'ssl' not in self.standard_library:
+                    continue
+
+                dep = dep[1:]
+            elif dep[0] == '!':
+                if 'ssl' in self.standard_library:
+                    continue
+
+                dep = dep[1:]
+
             self._set_dependency_state(all_modules, dep, visit,
                     (dep_state.explicit or dep_state.implicit))
 
@@ -268,7 +309,7 @@ class Project(QObject):
 
         cls._assert(version is not None, "Invalid 'version'.")
 
-        if version <= 3:
+        if version < cls.min_version:
             raise UserException("The project's format is no longer supported.")
 
         if version > cls.version:
@@ -279,13 +320,21 @@ class Project(QObject):
         project = cls()
         project._name = fi
 
+        # This was added in version 7.
+        project.using_default_locations = cls._get_bool(root,
+                'usingdefaultlocations', 'Project', default=False)
+
         # The Python specific configuration.
         python = root.find('Python')
         cls._assert(python is not None, "Missing 'Python' tag.")
 
         project.python_host_interpreter = python.get('hostinterpreter', '')
+
+        # This was added in version 5.
+        project.python_use_platform = cls._replace_scopes(
+                python.get('platformpython', '')).split()
+
         project.python_source_dir = python.get('sourcedir', '')
-        project.python_ssl = cls._get_bool(python, 'ssl', 'Python')
         project.python_target_include_dir = python.get('targetincludedir', '')
         project.python_target_library = python.get('targetlibrary', '')
         project.python_target_stdlib_dir = python.get('targetstdlibdir', '')
@@ -294,6 +343,10 @@ class Project(QObject):
         minor = cls._get_int(python, 'minor', 'Python')
         patch = cls._get_int(python, 'patch', 'Python', default=0)
         project.python_target_version = (major, minor, patch)
+        if project.python_target_version not in supported_python_versions:
+            raise UserException(
+                    "Python v{0}.{1}.{2} is not supported.".format(major,
+                            minor, patch))
 
         # The application specific configuration.
         application = root.find('Application')
@@ -304,12 +357,19 @@ class Project(QObject):
                 'Application')
         project.application_is_console = cls._get_bool(application,
                 'isconsole', 'Application')
+
         project.application_is_bundle = cls._get_bool(application, 'isbundle',
                 'Application')
         project.application_name = application.get('name', '')
         project.application_script = application.get('script', '')
         project.sys_path = application.get('syspath', '')
 
+        # Any qmake configuration. This was added in version 5.
+        qmake_configuration = application.find('QMakeConfiguration')
+
+        if qmake_configuration is not None:
+            project.qmake_configuration = qmake_configuration.text
+
         # Any application package.
         app_package = application.find('Package')
 
@@ -339,12 +399,23 @@ class Project(QObject):
             cls._assert(name is not None,
                     "Missing 'ExternalLib.name' attribute.")
 
-            defines = external_lib_element.get('defines', '')
-            includepath = external_lib_element.get('includepath', '')
-            libs = external_lib_element.get('libs', '')
+            defines = cls._fix_scopes(external_lib_element.get('defines', ''))
+            includepath = cls._fix_scopes(
+                    external_lib_element.get('includepath', ''))
+            libs = cls._fix_scopes(external_lib_element.get('libs', ''))
 
-            project.external_libraries.append(
-                    ExternalLibrary(name, defines, includepath, libs))
+            external_lib = ExternalLibrary(name, defines, includepath, libs)
+
+            target = external_lib_element.get('target')
+            if target is None:
+                # The project format is version 6 or earlier.
+                target_list = [p.name for p in Platform.all_platforms]
+            else:
+                target_list = [target]
+
+            for target in target_list:
+                project.external_libraries.setdefault(target, []).append(
+                        external_lib)
 
         # Any other Python packages.
         project.other_packages = [cls._load_package(package)
@@ -352,21 +423,29 @@ class Project(QObject):
 
         # Any other extension module.
         for extension_module_element in root.iterfind('ExtensionModule'):
-            name = extension_module_element.get('name')
+            name = cls._fix_scopes(extension_module_element.get('name'))
             cls._assert(name is not None,
                     "Missing 'ExtensionModule.name' attribute.")
 
-            libs = extension_module_element.get('libs')
-            cls._assert(path is not None,
-                    "Missing 'ExtensionModule.libs' attribute.")
+            qt = cls._fix_scopes(extension_module_element.get('qt', ''))
+            config = cls._fix_scopes(
+                    extension_module_element.get('config', ''))
+            sources = cls._fix_scopes(
+                    extension_module_element.get('sources', ''))
+            defines = cls._fix_scopes(
+                    extension_module_element.get('defines', ''))
+            includepath = cls._fix_scopes(
+                    extension_module_element.get('includepath', ''))
+            libs = cls._fix_scopes(extension_module_element.get('libs', ''))
 
-            project.other_extension_modules.append(ExtensionModule(name, libs))
+            project.other_extension_modules.append(
+                    ExtensionModule(name, qt, config, sources, defines,
+                            includepath, libs))
 
-        # The other configuration.
-        others = root.find('Others')
-        if others is not None:
-            project.build_dir = others.get('builddir', '')
-            project.qmake = others.get('qmake', '')
+        # If the default locations are being used then use the current defaults
+        # instead of those (possibly out of date) in the project file.
+        if project.using_default_locations:
+            project.set_default_locations()
 
         return project
 
@@ -434,14 +513,14 @@ class Project(QObject):
         return contents
 
     @classmethod
-    def _get_bool(cls, element, name, context):
+    def _get_bool(cls, element, name, context, default=None):
         """ Get a boolean attribute from an element. """
 
         value = element.get(name)
         try:
             value = int(value)
         except:
-            value = None
+            value = default
 
         cls._assert(value is not None,
                 "Missing or invalid boolean value of '{0}.{1}'.".format(
@@ -471,18 +550,23 @@ class Project(QObject):
         """
 
         root = Element('Project', attrib={
-            'version': str(self.version)})
+            'version': str(self.version),
+            'usingdefaultlocations': str(int(self.using_default_locations))})
 
-        SubElement(root, 'Python', attrib={
-            'hostinterpreter': self.python_host_interpreter,
-            'sourcedir': self.python_source_dir,
-            'ssl': str(int(self.python_ssl)),
-            'targetincludedir': self.python_target_include_dir,
-            'targetlibrary': self.python_target_library,
-            'targetstdlibdir': self.python_target_stdlib_dir,
+        attrib = {
+            'platformpython': ' '.join(self.python_use_platform),
             'major': str(self.python_target_version[0]),
             'minor': str(self.python_target_version[1]),
-            'patch': str(self.python_target_version[2])})
+            'patch': str(self.python_target_version[2])}
+
+        if not self.using_default_locations:
+            attrib['hostinterpreter'] = self.python_host_interpreter
+            attrib['sourcedir'] = self.python_source_dir
+            attrib['targetincludedir'] = self.python_target_include_dir
+            attrib['targetlibrary'] = self.python_target_library
+            attrib['targetstdlibdir'] = self.python_target_stdlib_dir
+
+        SubElement(root, 'Python', attrib=attrib)
 
         application = SubElement(root, 'Application', attrib={
             'entrypoint': self.application_entry_point,
@@ -493,7 +577,10 @@ class Project(QObject):
             'script': self.application_script,
             'syspath': self.sys_path})
 
-        if self.application_package.name != '':
+        if self.qmake_configuration != '':
+            SubElement(application, 'QMakeConfiguration').text = self.qmake_configuration
+
+        if self.application_package.name is not None:
             self._save_package(application, self.application_package)
 
         for pyqt_module in self.pyqt_modules:
@@ -504,12 +591,14 @@ class Project(QObject):
             SubElement(root, 'StdlibModule', attrib={
                 'name': stdlib_module})
 
-        for external_lib in self.external_libraries:
-            SubElement(root, 'ExternalLib', attrib={
-                'name': external_lib.name,
-                'defines': external_lib.defines,
-                'includepath': external_lib.includepath,
-                'libs': external_lib.libs})
+        for target, external_libs in self.external_libraries.items():
+            for external_lib in external_libs:
+                SubElement(root, 'ExternalLib', attrib={
+                    'target': target,
+                    'name': external_lib.name,
+                    'defines': external_lib.defines,
+                    'includepath': external_lib.includepath,
+                    'libs': external_lib.libs})
 
         for package in self.other_packages:
             self._save_package(root, package)
@@ -517,16 +606,17 @@ class Project(QObject):
         for extension_module in self.other_extension_modules:
             SubElement(root, 'ExtensionModule', attrib={
                 'name': extension_module.name,
+                'qt': extension_module.qt,
+                'config': extension_module.config,
+                'sources': extension_module.sources,
+                'defines': extension_module.defines,
+                'includepath': extension_module.includepath,
                 'libs': extension_module.libs})
 
-        SubElement(root, 'Others', attrib={
-            'builddir': self.build_dir,
-            'qmake': self.qmake})
-
         tree = ElementTree(root)
 
         try:
-            tree.write(QDir.toNativeSeparators(file_name), encoding='unicode',
+            tree.write(QDir.toNativeSeparators(file_name), encoding='utf-8',
                     xml_declaration=True)
         except Exception as e:
             raise UserException(
@@ -569,6 +659,38 @@ class Project(QObject):
         if not ok:
             raise UserException("The project file is invalid.", detail)
 
+    @classmethod
+    def _fix_scopes(cls, value):
+        """ In version 6 and earlier scopes where qmake scopes, starting with
+        version 7 they are our well defined platform names.  This handles the
+        conversion for a string.
+        """
+
+        if '#' in value:
+            updated = []
+
+            for single in value.split():
+                parts = single.split('#', maxsplit=1)
+                if len(parts) == 2:
+                    lhs, rhs = parts
+                    updated.append(cls._replace_scopes(lhs) + '#' + rhs)
+                else:
+                    updated.append(single)
+
+            value = ' '.join(updated)
+
+        return value
+
+    @staticmethod
+    def _replace_scopes(value):
+        """ Replace any qmake scopes in a value. """
+
+        value = value.replace('linux-*', 'linux')
+        value = value.replace('macx', 'macos')
+        value = value.replace('win32', 'win')
+
+        return value
+
 
 class QrcPackage():
     """ The encapsulation of a memory-filesystem package. """
@@ -576,9 +698,10 @@ class QrcPackage():
     def __init__(self):
         """ Initialise the package. """
 
-        self.name = ''
+        self.name = None
         self.contents = []
-        self.exclusions = ['*.pyc', '*.pyo', '__pycache__', '*-info']
+        self.exclusions = ['*.pyc', '*.pyd', '*.pyo', '*.pyx', '*.pxi',
+                '__pycache__', '*-info', 'EGG_INFO', '*.so']
 
     def copy(self):
         """ Return a copy of the package. """
@@ -642,10 +765,15 @@ class ExternalLibrary():
 class ExtensionModule():
     """ The encapsulation of an extension module. """
 
-    def __init__(self, name, libs):
+    def __init__(self, name, qt, config, sources, defines, includepath, libs):
         """ Initialise the extension module. """
 
         self.name = name
+        self.qt = qt
+        self.config = config
+        self.sources = sources
+        self.defines = defines
+        self.includepath = includepath
         self.libs = libs
 
 
