@@ -1511,7 +1511,7 @@ pyconfig = (
 
     # Define if you want to build the _decimal module using a coroutine-local
     # rather than a thread-local context.
-    Config('WITH_DECIMAL_CONTEXTVAR')
+    Config('WITH_DECIMAL_CONTEXTVAR'),
 
     # Define if you want documentation strings in extension modules
     Config('WITH_DOC_STRINGS'),
